using System;
using System.Collections;
using Server.Targeting;
using Server.Network;
using Server.Spells;
using Server.Items;
using Server.Mobiles;
using Server.ContextMenus;

namespace Server.Mobiles
{
	/// <summary>
	/// Summary description for MobileAI.
	/// </summary>
	/// 
	public enum FightMode
	{
		None,			// Never focus on others
		Agressor,		// Only attack agressors
		Strongest,		// Attack the strongest
		Weakest,		// Attack the weakest
		Closest 		// Attack the closest
	}

	public enum OrderType
	{
		None,			//When no order, let's roam
		Come,			//"(All/Name) come"  Summons all or one pet to your location.  
		Drop,			//"(Name) drop"  Drops its loot to the ground (if it carries any).  
		Follow,			//"(Name) follow"  Follows targeted being.  
						//"(All/Name) follow me"  Makes all or one pet follow you.  
		Friend,			//"(Name) friend"  Allows targeted player to confirm resurrection. 
		Guard,			//"(Name) guard"  Makes the specified pet guard you. Pets can only guard their owner. 
						//"(All/Name) guard me"  Makes all or one pet guard you.  
		Attack,			//"(All/Name) kill", 
						//"(All/Name) attack"  All or the specified pet(s) currently under your control attack the target. 
		Patrol,			//"(Name) patrol"  Roves between two or more guarded targets.  
		Release,		//"(Name) release"  Releases pet back into the wild (removes "tame" status). 
		Stay,			//"(All/Name) stay" All or the specified pet(s) will stop and stay in current spot. 
		Stop,			//"(All/Name) stop Cancels any current orders to attack, guard or follow.  
		Transfert		//"(Name) transfer" Transfers complete ownership to targeted player. 
	}

	public enum MeatType
	{
		Ribs,
		Bird
	}

	public enum HideType
	{
		Regular,
		Spined,
		Horned,
		Barbed
	}

	public class BaseCreature : Mobile
	{
		private BaseAI	m_AI;					// THE AI
		
		private AIType	m_CurrentAI;			// The current AI
		private AIType	m_DefaultAI;			// The default AI

		private Mobile	m_FocusMob;				// Use focus mob instead of combatant, maybe we don't whan to fight
		private FightMode m_FightMode;			// The style the mob uses

		private int		m_iRangePerception;		// The view area
		private int		m_iRangeFight;			// The fight distance
       
		private bool	m_bDebugAI;				// Show debug AI messages

		private int		m_iTeam;				// Monster Team

        private double	m_dActiveSpeed;			// Timer speed when active
		private double	m_dPassiveSpeed;		// Timer speed when not active
		private double	m_dCurrentSpeed;		// The current speed, lets say it could be changed by something;

		private Point3D m_pHome;				// The home position of the creature, used by some AI
		private int		m_iRangeHome = 10;		// The home range of the creature

		ArrayList		m_arSpellAttack;		// List of attack spell/power
		ArrayList		m_arSpellDefense;		// Liste of defensive spell/power

		private bool		m_bWaitForTarget = false;

		private bool		m_bControled;		// Is controled
		private Mobile		m_ControlMaster;	// My master
		private Mobile		m_ControlTarget;	// My target mobile
		private Point3D		m_ControlDest;		// My target destination (patrol)
		private OrderType	m_ControlOrder;		// My order

		private double	m_dMinTameSkill;
		private double	m_dMaxTameSkill;
		private bool	m_bTamable;

		private bool		m_bSummoned = false;
		private DateTime	m_SummonEnd;
		private int			m_iControlSlots = 1;

		private bool		m_bBardProvoked = false;
		private bool		m_bBardPacified = false;
		private Mobile		m_bBardMaster = null;
		private Mobile		m_bBardTarget = null;
		private DateTime	m_timeBardEnd;

		[CommandProperty( AccessLevel.GameMaster )]
		public override int HitsMax
		{
			get
			{
				return Str;
			}
		}

		public virtual void OnCarve( Mobile from, Corpse corpse )
		{
			int feathers = Feathers;
			int wool = Wool;
			int meat = Meat;
			int hides = Hides;

			if ( (feathers == 0 && wool == 0 && meat == 0 && hides == 0) || Summoned )
			{
				from.SendLocalizedMessage( 500485 ); // You see nothing useful to carve from the corpse.
			}
			else
			{
				new Blood( 0x122D ).MoveToWorld( corpse.Location, corpse.Map );

				if ( feathers != 0 )
				{
					corpse.DropItem( new Feather( feathers ) );
					from.SendLocalizedMessage( 500479 ); // You pluck the bird. The feathers are now on the corpse.
				}

				if ( wool != 0 )
				{
					corpse.DropItem( new Wool( wool ) );
					from.SendLocalizedMessage( 500483 ); // You shear it, and the wool is now on the corpse.
				}

				if ( meat != 0 )
				{
					if ( MeatType == MeatType.Ribs )
						corpse.DropItem( new RawRibs( meat ) );
					else if ( MeatType == MeatType.Bird )
						corpse.DropItem( new RawBird( meat ) );

					from.SendLocalizedMessage( 500467 ); // You carve some meat, which remains on the corpse.
				}

				if ( hides != 0 )
				{
					if ( HideType == HideType.Regular )
						corpse.DropItem( new Hides( hides ) );
					else if ( HideType == HideType.Spined )
						corpse.DropItem( new SpinedHides( hides ) );
					else if ( HideType == HideType.Horned )
						corpse.DropItem( new HornedHides( hides ) );
					else if ( HideType == HideType.Barbed )
						corpse.DropItem( new BarbedHides( hides ) );

					from.SendLocalizedMessage( 500471 ); // You skin it, and the hides are now in the corpse.
				}

				corpse.Carved = true;

				if ( Notoriety.Compute( from, corpse ) == Notoriety.Innocent )
					from.CriminalAction( true );
			}
		}

		public BaseCreature(AIType ai,
							FightMode mode,
							int iRangePerception,
							int iRangeFight,
							double dActiveSpeed, 
							double dPassiveSpeed)
		{
			m_CurrentAI = ai;
			m_DefaultAI = ai;

			m_iRangePerception = iRangePerception;
			m_iRangeFight = iRangeFight;
			
			m_FightMode = mode;

			m_iTeam = 0;

			m_dActiveSpeed = dActiveSpeed;
			m_dPassiveSpeed = dPassiveSpeed;
			m_dCurrentSpeed = dPassiveSpeed;

			m_bDebugAI = false;

			m_arSpellAttack = new ArrayList();
			m_arSpellDefense = new ArrayList();

			m_bControled = false;
			m_ControlMaster = null;
			m_ControlTarget = null;
			m_ControlOrder = OrderType.None;

			m_bTamable = false;

			ChangeAIType(AI);
		}

		public BaseCreature( Serial serial ) : base( serial )
		{
			m_arSpellAttack = new ArrayList();
			m_arSpellDefense = new ArrayList();

			m_bDebugAI = false;
		}

		public override void Serialize( GenericWriter writer )
		{
			base.Serialize( writer );

			writer.Write( (int) 2 ); // version

			writer.Write( (int)m_CurrentAI );
			writer.Write( (int)m_DefaultAI );

			writer.Write( (int)m_iRangePerception );
			writer.Write( (int)m_iRangeFight );

			writer.Write( (int)m_iTeam );

			writer.Write( (double)m_dActiveSpeed );
			writer.Write( (double)m_dPassiveSpeed );
			writer.Write( (double)m_dCurrentSpeed );

			writer.Write( (int) m_pHome.X );
			writer.Write( (int) m_pHome.Y );
			writer.Write( (int) m_pHome.Z );

			// Version 1
			writer.Write( (int) m_iRangeHome );

			int i=0;

			writer.Write( (int) m_arSpellAttack.Count );
			for ( i=0; i< m_arSpellAttack.Count; i++ )
			{
				writer.Write( m_arSpellAttack[i].ToString() );
			}

			writer.Write( (int) m_arSpellDefense.Count );
			for ( i=0; i< m_arSpellDefense.Count; i++ )
			{
				writer.Write( m_arSpellDefense[i].ToString() );
			}

			// Version 2
			writer.Write( (int) m_FightMode );

			writer.Write( (bool) m_bControled );
			writer.Write( (Mobile) m_ControlMaster );
			writer.Write( (Mobile) m_ControlTarget );
			writer.Write( (Point3D) m_ControlDest );
			writer.Write( (int) m_ControlOrder );
			writer.Write( (double) m_dMinTameSkill );
			writer.Write( (double) m_dMaxTameSkill );
			writer.Write( (bool) m_bTamable );
			writer.Write( (bool) m_bSummoned );

			if ( m_bSummoned )
			{
				writer.Write( m_SummonEnd - DateTime.Now );
			}

			writer.Write( (int) m_iControlSlots );
		}

		public override void Deserialize( GenericReader reader )
		{
			base.Deserialize( reader );

			int version = reader.ReadInt();

			m_CurrentAI = (AIType)reader.ReadInt();
			m_DefaultAI = (AIType)reader.ReadInt();

			m_iRangePerception = reader.ReadInt();
			m_iRangeFight = reader.ReadInt();

			m_iTeam = reader.ReadInt();

			m_dActiveSpeed = reader.ReadDouble();
			m_dPassiveSpeed = reader.ReadDouble();
			m_dCurrentSpeed = reader.ReadDouble();

			m_pHome.X = reader.ReadInt();
			m_pHome.Y = reader.ReadInt();
			m_pHome.Z = reader.ReadInt();

			if ( version >= 1 )
			{
				m_iRangeHome = reader.ReadInt();

				int i, iCount;
				
				iCount = reader.ReadInt();
				for ( i=0; i< iCount; i++ )
				{
					string str = reader.ReadString();
					Type type = Type.GetType( str );

					if ( type != null )
					{
						m_arSpellAttack.Add( type );
					}
				}

				iCount = reader.ReadInt();
				for ( i=0; i< iCount; i++ )
				{
					string str = reader.ReadString();
					Type type = Type.GetType( str );

					if ( type != null )
					{
						m_arSpellDefense.Add( type );
					}			
				}
			}
			else
			{
				m_iRangeHome = 0;
			}

			if ( version >= 2 )
			{
				m_FightMode = ( FightMode )reader.ReadInt();

				m_bControled = reader.ReadBool();
				m_ControlMaster = reader.ReadMobile();
				m_ControlTarget = reader.ReadMobile();
				m_ControlDest = reader.ReadPoint3D();
				m_ControlOrder = (OrderType) reader.ReadInt();

				m_dMinTameSkill = reader.ReadDouble();
				m_dMaxTameSkill = reader.ReadDouble();
				m_bTamable = reader.ReadBool();
				m_bSummoned = reader.ReadBool();

				if ( m_bSummoned )
				{
					TimeSpan duration = reader.ReadTimeSpan();

					new UnsummonTimer( m_ControlMaster, this, duration ).Start();
					m_SummonEnd = DateTime.Now + duration;
				}

				m_iControlSlots = reader.ReadInt();
			}
			else
			{
				m_FightMode = FightMode.Closest;

				m_bControled = false;
				m_ControlMaster = null;
				m_ControlTarget = null;
				m_ControlOrder = OrderType.None;
			}

			ChangeAIType(m_CurrentAI);
		}

		public virtual void ChangeAIType(AIType NewAI)
		{
			if ( m_AI != null )
			{
				m_AI.m_Timer.Stop();
			}
			m_AI = null;

			switch (NewAI)
			{
				case AIType.AI_Melee:
					m_AI = new MeleeAI(this);
					break;
				case AIType.AI_Animal:
					m_AI = new AnimalAI(this);
					break;
				case AIType.AI_Berserk:
					m_AI = new BerserkAI(this);
					break;
				case AIType.AI_Archer:
					m_AI = new ArcherAI(this);
					break;
				case AIType.AI_Healer:
					m_AI = new HealerAI(this);
					break;
				case AIType.AI_Vendor:
					m_AI = new VendorAI(this);
					break;
				case AIType.AI_Mage:
					m_AI = new MageAI(this);
					break;
				case AIType.AI_Predator:
					m_AI = new PredatorAI(this);
					break;
				case AIType.AI_Thief:
					m_AI = new ThiefAI(this);
					break;
			}
		}

		public virtual void ChangeAIToDefault()
		{
           ChangeAIType(m_DefaultAI);
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public AIType AI
		{
			get
			{
				return m_CurrentAI;
			}
			set
			{
				m_CurrentAI = value;

				if (m_CurrentAI == AIType.AI_Use_Default)
				{
					m_CurrentAI = m_DefaultAI;
				}
				
				ChangeAIType(m_CurrentAI);
			}
		}

		[CommandProperty( AccessLevel.Administrator )]
		public bool Debug
		{
			get
			{
				return m_bDebugAI;
			}
			set
			{
				m_bDebugAI = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public int Team
		{
			get
			{
				return m_iTeam;
			}
			set
			{
				m_iTeam = value;
				
				OnTeamChange();
			}
		}

		public virtual void OnTeamChange()
		{
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Mobile FocusMob
		{
			get
			{
				return m_FocusMob;
			}
			set
			{
				m_FocusMob = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public FightMode FightMode
		{
			get
			{
				return m_FightMode;
			}
			set
			{
				m_FightMode = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public int RangePerception
		{
			get
			{
				return m_iRangePerception;
			}
			set
			{
				m_iRangePerception = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public int RangeFight
		{
			get
			{
				return m_iRangeFight;
			}
			set
			{
				m_iRangeFight = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public int RangeHome
		{
			get
			{
				return m_iRangeHome;
			}
			set
			{
				m_iRangeHome = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public double ActiveSpeed
		{
			get
			{
				return m_dActiveSpeed;
			}
			set
			{
				m_dActiveSpeed = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public double PassiveSpeed
		{
			get
			{
				return m_dPassiveSpeed;
			}
			set
			{
				m_dPassiveSpeed = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public double CurrentSpeed
		{
			get
			{
				return m_dCurrentSpeed;
			}
			set
			{
				if ( m_dCurrentSpeed != value )
				{
					m_dCurrentSpeed = value;

					if (m_AI != null)
					{
						m_AI.OnCurrentSpeedChanged();
					}
				}
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Point3D Home
		{
			get
			{
				return m_pHome;
			}
			set
			{
				m_pHome = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public bool Controled
		{
			get
			{
				return m_bControled;
			}
			set
			{
				if ( m_bControled == value )
					return;

				m_bControled = value;
				Delta( MobileDelta.Noto );
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Mobile ControlMaster
		{
			get
			{
				return m_ControlMaster;
			}
			set
			{
				if ( m_ControlMaster == value )
					return;

				m_ControlMaster = value;
				Delta( MobileDelta.Noto );
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Mobile ControlTarget
		{
			get
			{
				return m_ControlTarget;
			}
			set
			{
				m_ControlTarget = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Point3D ControlDest
		{
			get
			{
				return m_ControlDest;
			}
			set
			{
				m_ControlDest = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public OrderType ControlOrder
		{
			get
			{
				return m_ControlOrder;
			}
			set
			{
				m_ControlOrder = value;

				m_AI.OnCurrentOrderChanged();
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public bool BardProvoked
		{
			get
			{
				return m_bBardProvoked;
			}
			set
			{
				m_bBardProvoked = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public bool BardPacified
		{
			get
			{
				return m_bBardPacified;
			}
			set
			{
				m_bBardPacified = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Mobile BardMaster
		{
			get
			{
				return m_bBardMaster;
			}
			set
			{
				m_bBardMaster = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public Mobile BardTarget
		{
			get
			{
				return m_bBardTarget;
			}
			set
			{
				m_bBardTarget = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public DateTime BardEndTime
		{
			get
			{
				return m_timeBardEnd;
			}
			set
			{
				m_timeBardEnd = value;
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public double MinTameSkill
		{
			get
			{
				return m_dMinTameSkill;
			}
			set
			{
				m_dMinTameSkill = value;
				if ( m_dMaxTameSkill < value )
				{
					m_dMaxTameSkill = value + 20;
				}
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public double MaxTameSkill
		{
			get
			{
				return m_dMaxTameSkill;
			}
			set
			{
				m_dMaxTameSkill = value;
				if ( m_dMinTameSkill > value )
				{
					m_dMinTameSkill = value - 20;
				}
			}
		}

		[CommandProperty( AccessLevel.GameMaster )]
		public bool Tamable
		{
			get
			{
				return m_bTamable;
			}
			set
			{
				m_bTamable = value;
			}
		}

		[CommandProperty( AccessLevel.Administrator )]
		public bool WaitForTarget
		{
			get
			{
				return m_bWaitForTarget;
			}
			set
			{
				m_bWaitForTarget = value;
			}
		}

		[CommandProperty( AccessLevel.Administrator )]
		public bool Summoned
		{
			get
			{
				return m_bSummoned;
			}
			set
			{
				if ( m_bSummoned == value )
					return;

				m_bSummoned = value;
				Delta( MobileDelta.Noto );
			}
		}

		[CommandProperty( AccessLevel.Administrator )]
		public int ControlSlots
		{
			get
			{
				return m_iControlSlots;
			}
			set
			{
				m_iControlSlots = value;
			}
		}

		public virtual int Feathers{ get{ return 0; } }
		public virtual int Wool{ get{ return 0; } }

		public virtual MeatType MeatType{ get{ return MeatType.Ribs; } }
		public virtual int Meat{ get{ return 0; } }

		public virtual int Hides{ get{ return 0; } }
		public virtual HideType HideType{ get{ return HideType.Regular; } }

		public override void OnAfterDelete()
		{
			if ( m_AI != null )
			{
				if ( m_AI.m_Timer != null )
					m_AI.m_Timer.Stop();

				m_AI = null;
			}

			FocusMob = null;

			base.OnAfterDelete();
		}

		public void DebugSay( string text )
		{
			if ( m_bDebugAI )
				this.PublicOverheadMessage( MessageType.Regular, 41, false, text );
		}

		public void DebugSay( string format, params object[] args )
		{
			if ( m_bDebugAI )
				this.PublicOverheadMessage( MessageType.Regular, 41, false, String.Format( format, args ) );
		}

		/*
		 * Will need to be givent a better name
		 * 
		 * This function can be overriden.. so a "Strongest" mobile, can have a different definition depending
		 * on who check for value
		 * -Could add a FightMode.Prefered
		 * 
		 */
		public virtual double GetValueFrom( Mobile m, FightMode acqType, bool bPlayerOnly )
		{
			if ( ( bPlayerOnly && m.Player ) ||  !bPlayerOnly )
			{
				switch( acqType )
				{
					case FightMode.Strongest : 
						return (m.Skills[SkillName.Tactics].Value + m.Str); //returns strongest mobile

					case FightMode.Weakest : 
						return -m.Hits; // returns weakest mobile

					default : 
						return -GetDistanceToSqrt( m ); // returns closest mobile
				}
			}
			else
			{
				return double.MinValue;
			}
		}

		// Turn, - for let, + for right
		// Basic for now, need works
		public virtual void Turn(int iTurnSteps)
		{
			int v = (int)Direction;

			Direction = (Direction)((((v & 0x7) + iTurnSteps) & 0x7) | (v & 0x80));
		}

		public bool IsHurt()
		{
			return ( Hits != HitsMax );
		}

		public double GetHomeDistance()
		{
            return GetDistanceToSqrt( m_pHome );
		}

		public virtual int GetTeamSize(int iRange)
		{
			int iCount = 0;
			Map map = this.Map;

			if ( map != null )
			{
				IPooledEnumerable eable = map.GetMobilesInRange( Location, iRange);

				foreach ( Mobile m in eable )
				{
					if (m is BaseCreature)
					{
						if ( ((BaseCreature)m).Team == Team )
						{
							if ( !m.Deleted )
							{
								if ( m != this )
								{
									if ( CanSee( m ) )
									{
										iCount++;
									}
								}
							}
						}
					}
				}

				eable.Free();
			}
			
			return iCount;
		}

		// Do my combatant is attaking me??
		public bool IsCombatantAnAgressor()
		{
			if (Combatant != null)
			{
				if (Combatant.Combatant == this)
				{
					return true;
				}
			}
			return false;
		}

		private class TameEntry : ContextMenuEntry
		{
			private BaseCreature m_Mobile;

			public TameEntry( BaseCreature mobile ) : base( 6130, 6 )
			{
				m_Mobile = mobile;
			}

			public override void OnClick()
			{
				Owner.From.TargetLocked = true;

				if ( Owner.From.UseSkill( SkillName.AnimalTaming ) )
					Owner.From.Target.Invoke( Owner.From, m_Mobile );

				Owner.From.TargetLocked = false;
			}
		}

		public virtual void BeginTeach( SkillName skill, Mobile m )
		{
			Skill ourSkill = Skills[skill];
			Skill theirSkill = m.Skills[skill];

			double toTeach = ourSkill.Base / 2.0;

			if ( toTeach > 30.0 )
				toTeach = 30.0;
			else if ( toTeach < 10.0 )
				return;

			if ( theirSkill.Lock != SkillLock.Up || (m.Skills.Total - theirSkill.BaseFixedPoint + (int)(toTeach * 10)) > m.Skills.Cap )
			{
				m.SendLocalizedMessage( 501510 ); // Make sure this skill is marked to raise. If you are near the skill cap (700 points) you may need to lose some points in another skill first.
			}
			else if ( theirSkill.Base > toTeach )
			{
				Say( 501508 ); // I cannot teach thee, for thou knowest more than I!
			}
			else if ( theirSkill.Base == toTeach )
			{
				Say( 501509 ); // I cannot teach thee, for thou knowest all I can teach!
			}
			else
			{
				Container pack = m.Backpack;

				int cost = (int)((toTeach - theirSkill.Base) * 10);

				if ( pack == null || !pack.ConsumeTotal( typeof( Gold ), cost ) )
				{
					Say( 500683 ); // I am sorry, but thou lacks the gold for training.
				}
				else
				{
					Say( 501539 ); // Let me show thee something of how this is done.
					m.SendLocalizedMessage( 501540 ); // Your skill level increases.

					theirSkill.Base = toTeach;
				}
			}
		}

		public override void GetContextMenuEntries( Mobile from, ArrayList list )
		{
			base.GetContextMenuEntries( from, list );

			if ( m_AI != null )
				m_AI.GetContextMenuEntries( from, list );

			if ( m_bTamable && !m_bControled )
				list.Add( new TameEntry( this ) );

			if ( Body.IsHuman )
			{
				for ( int i = 0; i < 50; ++i )
				{
					Skill skill = Skills[i];
					Skill theirSkill = from.Skills[i];

					if ( skill != null && theirSkill != null && skill.Base >= 20.0 )
					{
						double toTeach = skill.Base / 2.0;

						if ( toTeach > 30.0 )
							toTeach = 30.0;

						if ( toTeach > theirSkill.Base )
							list.Add( new TeachEntry( (SkillName)i, this, from ) );
					}
				}
			}
		}

		public override void OnSpeech( SpeechEventArgs e )
		{
			base.OnSpeech( e );

			if ( !e.Handled && e.Mobile.InRange( this, m_iRangePerception ) )
			{
				if ( m_AI != null )
				{
					m_AI.OnSpeech( e );
				}
			}
		}

		public void AddSpellAttack( Type type )
		{
			m_arSpellAttack.Add ( type );
		}

		public void AddSpellDefense( Type type )
		{
			m_arSpellDefense.Add ( type );
		}

		public Spell GetAttackSpellRandom()
		{
			if ( m_arSpellAttack.Count > 0 )
			{
				Type type = (Type) m_arSpellAttack[Utility.Random(m_arSpellAttack.Count)];

				object[] args = {this, null};
				return Activator.CreateInstance( type, args ) as Spell;
			}
			else
			{
				return null;
			}
		}

		public Spell GetDefenseSpellRandom()
		{
			if ( m_arSpellDefense.Count > 0 )
			{
				Type type = (Type) m_arSpellDefense[Utility.Random(m_arSpellDefense.Count)];

				object[] args = {this, null};
				return Activator.CreateInstance( type, args ) as Spell;
			}
			else
			{
				return null;
			}
		}

		public Spell GetSpellSpecific( Type type )
		{
			int i;

			for ( i=0; i< m_arSpellAttack.Count; i++ )
			{
				if ( m_arSpellAttack[i] == type )
				{
					object[] args = {this, null};
					return Activator.CreateInstance( type, args ) as Spell;
				}
			}

			for ( i=0; i< m_arSpellDefense.Count; i++ )
			{
				if ( m_arSpellDefense[i] == type )
				{
					object[] args = {this, null};
					return Activator.CreateInstance( type, args ) as Spell;
				}			
			}

			return null;
		}


		public void SetStr( int val )
		{
			RawStr = val;
			Hits = HitsMax;
		}

		public void SetStr( int min, int max )
		{
			RawStr = Utility.RandomMinMax( min, max );
			Hits = HitsMax;
		}

		public void SetDex( int val )
		{
			RawDex = val;
			Stam = StamMax;
		}

		public void SetDex( int min, int max )
		{
			RawDex = Utility.RandomMinMax( min, max );
			Stam = StamMax;
		}

		public void SetInt( int val )
		{
			RawInt = val;
			Mana = ManaMax;
		}

		public void SetInt( int min, int max )
		{
			RawInt = Utility.RandomMinMax( min, max );
			Mana = ManaMax;
		}

		public void SetSkill( SkillName name, double val )
		{
			Skills[name].BaseFixedPoint = (int)(val * 10);
		}

		public void SetSkill( SkillName name, double min, double max )
		{
			int minFixed = (int)(min * 10);
			int maxFixed = (int)(max * 10);

			Skills[name].BaseFixedPoint = Utility.RandomMinMax( minFixed, maxFixed );
		}

		public void SetFameLevel( int level )
		{
			switch ( level )
			{
				case 1: Fame = Utility.RandomMinMax(     0,  1249 ); break;
				case 2: Fame = Utility.RandomMinMax(  1250,  2499 ); break;
				case 3: Fame = Utility.RandomMinMax(  2500,  4999 ); break;
				case 4: Fame = Utility.RandomMinMax(  5000,  9999 ); break;
				case 5: Fame = Utility.RandomMinMax( 10000, 10000 ); break;
			}
		}

		public void SetKarmaLevel( int level )
		{
			switch ( level )
			{
				case 1: Karma = -Utility.RandomMinMax(   625,  1249 ); break;
				case 2: Karma = -Utility.RandomMinMax(  1250,  2499 ); break;
				case 3: Karma = -Utility.RandomMinMax(  2500,  4999 ); break;
				case 4: Karma = -Utility.RandomMinMax(  5000,  9999 ); break;
				case 5: Karma = -Utility.RandomMinMax( 10000, 10000 ); break;
			}
		}

		public void PackGem()
		{
			PackGem( 1 );
		}

		public void PackGem( int min, int max )
		{
			PackGem( Utility.RandomMinMax( min, max ) );
		}

		public void PackGem( int amount )
		{
			switch ( Utility.Random( 9 ) ) 
			{ 
				case 0: PackItem( new Amber( amount ) ); break;
				case 1: PackItem( new Amethyst( amount ) ); break;
				case 2: PackItem( new Citrine( amount ) ); break;
				case 3: PackItem( new Diamond( amount ) ); break;
				case 4: PackItem( new Emerald( amount ) ); break;
				case 5: PackItem( new Ruby( amount ) ); break;
				case 6: PackItem( new Sapphire( amount ) ); break;
				case 7: PackItem( new StarSapphire( amount ) ); break;
				case 8: PackItem( new Tourmaline( amount ) ); break;
			} 
		}

		public void PackReg( int min, int max )
		{
			PackReg( Utility.RandomMinMax( min, max ) );
		}

		public void PackReg( int amount )
		{
			switch ( Utility.Random( 8 ) )
			{
				case 0: PackItem( new BlackPearl( amount ) ); break;
				case 1: PackItem( new Bloodmoss( amount ) ); break;
				case 2: PackItem( new Garlic( amount ) ); break;
				case 3: PackItem( new Ginseng( amount ) ); break;
				case 4: PackItem( new MandrakeRoot( amount ) ); break;
				case 5: PackItem( new Nightshade( amount ) ); break;
				case 6: PackItem( new SulfurousAsh( amount ) ); break;
				case 7: PackItem( new SpidersSilk( amount ) ); break;
			}
		}

		public void PackItem( Item item )
		{
			if ( Summoned )
				return;

			Container pack = Backpack;

			if ( pack == null )
			{
				pack = new Backpack();

				pack.Movable = false;

				AddItem( pack );
			}

			pack.DropItem( item );
		}

		public override void OnSingleClick( Mobile from )
		{
			if ( Controled )
				PrivateOverheadMessage( MessageType.Regular, 0x3B2, Summoned ? 1049646 : 502006, from.NetState ); // (summoned)/(tame)

			base.OnSingleClick( from );
		}

		public override void OnDeath( Container c )
		{
			int totalFame = Fame / 100;
			int totalKarma = -Karma / 100;

			ArrayList toGive = new ArrayList();

			ArrayList list = Aggressors;
			for ( int i = 0; i < list.Count; ++i )
			{
				AggressorInfo info = (AggressorInfo)list[i];

				if ( (DateTime.Now - info.LastCombatTime) < TimeSpan.FromSeconds( 30.0 ) && !toGive.Contains( info.Attacker ) )
					toGive.Add( info.Attacker );
			}

			list = Aggressed;
			for ( int i = 0; i < list.Count; ++i )
			{
				AggressorInfo info = (AggressorInfo)list[i];

				if ( (DateTime.Now - info.LastCombatTime) < TimeSpan.FromSeconds( 30.0 ) && !toGive.Contains( info.Defender ) )
					toGive.Add( info.Defender );
			}

			for ( int i = 0; i < toGive.Count; ++i )
			{
				Mobile m = (Mobile)toGive[i];

				int karmaToGive = totalKarma;
				int fameToGive = totalFame;

				if ( (m.Karma + karmaToGive) > 10000 )
					karmaToGive = 10000 - m.Karma;
				else if ( (m.Karma + karmaToGive) < -10000 )
					karmaToGive = -10000 - m.Karma;

				if ( (m.Fame + fameToGive) > 10000 )
					fameToGive = 10000 - m.Fame;
				else if ( (m.Fame + fameToGive) < 0 )
					fameToGive = -m.Fame;

				m.Karma += karmaToGive;
				m.Fame += fameToGive;

				if ( fameToGive >= 40 )
					m.SendLocalizedMessage( 1019054 ); // You have gained a lot of fame.
				else if ( fameToGive >= 25 )
					m.SendLocalizedMessage( 1019053 ); // You have gained a good amount of fame.
				else if ( fameToGive >= 10 )
					m.SendLocalizedMessage( 1019052 ); // You have gained some fame.
				else if ( fameToGive >= 1 )
					m.SendLocalizedMessage( 1019051 ); // You have gained a little fame.
				else if ( fameToGive <= -40 )
					m.SendLocalizedMessage( 1019058 ); // You have lost a lot of fame.
				else if ( fameToGive <= -25 )
					m.SendLocalizedMessage( 1019057 ); // You have lost a good amount of fame.
				else if ( fameToGive <= -10 )
					m.SendLocalizedMessage( 1019056 ); // You have lost some fame.
				else if ( fameToGive <= -1 )
					m.SendLocalizedMessage( 1019055 ); // You have lost a little fame.

				if ( karmaToGive >= 40 )
					m.SendLocalizedMessage( 1019062 ); // You have gained a lot of karma.
				else if ( karmaToGive >= 25 )
					m.SendLocalizedMessage( 1019060 ); // You have gained some karma.
				else if ( karmaToGive >= 10 )
					m.SendLocalizedMessage( 1019060 ); // You have gained some karma.
				else if ( karmaToGive >= 1 )
					m.SendLocalizedMessage( 1019059 ); // You have gained a little karma.
				else if ( karmaToGive <= -40 )
					m.SendLocalizedMessage( 1019066 ); // You have lost a lot of karma.
				else if ( karmaToGive <= -25 )
					m.SendLocalizedMessage( 1019065 ); // You have lost a good amount of karma.
				else if ( karmaToGive <= -10 )
					m.SendLocalizedMessage( 1019064 ); // You have lost some karma.
				else if ( karmaToGive <= -1 )
					m.SendLocalizedMessage( 1019063 ); // You have lost a little karma.
			}

			base.OnDeath( c );

			if ( Summoned )
				c.Delete();
		}

		public override void OnDelete()
		{
			if ( Summoned || Controled )
			{
				if ( ControlMaster != null )
					ControlMaster.Followers -= ControlSlots;
			}

			base.OnDelete();
		}

		public override bool CanBeRenamedBy( Mobile from )
		{
			bool ret = base.CanBeRenamedBy( from );

			if ( Controled && from == ControlMaster )
				ret = true;

			return ret;
		}

		public void SetControlMaster( Mobile m )
		{
			if ( m == null )
			{
				if ( ControlMaster != null )
					ControlMaster.Followers -= ControlSlots;

				ControlMaster = null;
				Controled = false;
				ControlOrder = OrderType.None;
				Guild = null;
			}
			else
			{
				if ( m.Followers + ControlSlots > m.FollowersMax )
				{
					m.SendLocalizedMessage( 1049607 ); // You have too many followers to control that creature.
					return;
				}

				if ( ControlMaster != null )
					ControlMaster.Followers -= ControlSlots;
			
				ControlMaster = m;
				Controled = true;
				ControlOrder = OrderType.Come;
				ControlMaster.Followers += ControlSlots;
				Guild = null;

				Delta( MobileDelta.Noto );
			}
		}

		public static bool Summon( BaseCreature creature, Mobile caster, Point3D p, int sound, TimeSpan duration )
		{
			return Summon( creature, true, caster, p, sound, duration );
		}

		public static bool Summon( BaseCreature creature, bool controled, Mobile caster, Point3D p, int sound, TimeSpan duration )
		{
			if ( controled )
			{
				if ( caster.Followers + creature.ControlSlots > caster.FollowersMax )
				{
					caster.SendLocalizedMessage( 1049645 ); // You have too many followers to summon that creature.
					creature.Delete();
					return false;
				}
				creature.SetControlMaster( caster );
			}
			
			creature.Map = caster.Map;
			creature.Location = p;

			creature.RangeHome = 10;
			creature.Summoned = true;

			new UnsummonTimer( caster, creature, duration ).Start();
			creature.m_SummonEnd = DateTime.Now + duration;
			Effects.PlaySound( p, creature.Map, sound );

			return true;
		}

		public void Pacify( Mobile master, DateTime endtime )
		{
			BardPacified = true;
			BardEndTime = endtime;
		}
 
		public void Provoke( Mobile master, Mobile target, bool bSuccess )
		{
			BardProvoked = true;

			this.PublicOverheadMessage( MessageType.Emote, EmoteHue, false, "*looks furious*" );
 
			if ( bSuccess )
			{
				PlaySound( GetIdleSound() );
 
				BardMaster = master;
				BardTarget = target;
			}
			else
			{
				PlaySound( GetAngerSound() );

				BardMaster = master;
				BardTarget = target;
			}
		}

		public bool FindMyName( string str, bool bWithAll )
		{
			int i, j;
 
			if( str.Length < Name.Length )
			{
				return false;
			}
 
			string[] wordsString = str.Split(' ');
			string[] wordsName = Name.Split(' ');
 
			for ( j=0 ; j < wordsName.Length; j++ )
			{
				string wordName = wordsName[j];
 
				bool bFound = false;
				for ( i=0 ; i < wordsString.Length; i++ )
				{
					string word = wordsString[i];

					if ( word.ToLower() == wordName.ToLower())
					{
						bFound = true;
					}
 
					if ( bWithAll && word.ToLower() == "all" )
					{
						return true;
					}
				}
 
				if ( !bFound )
				{
					return false;
				}
			}
 
			return true;
		}

		public static void TeleportPets( Mobile master, Point3D loc, Map map )
		{
			if ( master.Map == null )
				return;

			ArrayList move = new ArrayList();

			IPooledEnumerable eable = master.Map.GetMobilesInRange( master.Location, 3 );

			foreach ( Mobile m in eable )
			{
				if ( m is BaseCreature )
				{
					BaseCreature pet = (BaseCreature)m;

					if ( pet.Controled && pet.ControlMaster == master )
					{
						if ( pet.ControlOrder == OrderType.Guard || pet.ControlOrder == OrderType.Follow || pet.ControlOrder == OrderType.Come )
							move.Add( pet );
					}
				}
			}

			eable.Free();

			foreach ( Mobile m in move )
			{
				m.Map = map;
				m.Location = loc;
			}
		}
	}
}