using System;
using System.Collections;
using Server.Items;
using Server.Network;
using Server.ContextMenus;
using Server.Mobiles;
using Server.Misc;

namespace Server.Mobiles
{
	public abstract class BaseVendor : BaseCreature, IVendor
	{
		private const int MaxSell = 5;


		protected abstract ArrayList SBInfos{ get; }

		private ArrayList m_ArmorBuyInfo = new ArrayList();
		private ArrayList m_ArmorSellInfo = new ArrayList();

		private DateTime m_LastRestock;

		public BaseVendor( string title ) : base( AIType.AI_Vendor, FightMode.None, 2, 1, 0.5, 2 )
		{
			LoadSBInfo();

			this.Title = title;
			InitBody();
			InitOutfit();

			Container pack;
			//these packs MUST exist, or the client will crash when the packets are sent
			pack = new Container( 0x9B2, 0x3C, 0x48, new Rectangle2D( 44, 65, 142, 94 ) );//stuff for a regular backpack
			pack.Layer = Layer.ShopBuy;
			pack.Movable = false;
			AddItem( pack );

			pack = new Container( 0x9B2, 0x3C, 0x48, new Rectangle2D( 44, 65, 142, 94 ) );//stuff for a regular backpack
			pack.Layer = Layer.ShopResale;
			pack.Movable = false;
			AddItem( pack );

			m_LastRestock = DateTime.Now;
		}
		
		public BaseVendor( Serial serial ) : base( serial )
		{
		}

		public DateTime LastRestock
		{
			get
			{
				return m_LastRestock;
			}
			set
			{
				m_LastRestock = value;
			}
		}

		public virtual TimeSpan RestockDelay
		{
			get
			{
				return TimeSpan.FromHours( 1 );
			}
		}

		public Container BuyPack
		{
			get
			{
				Container pack = FindItemOnLayer( Layer.ShopBuy ) as Container;
				if ( pack == null )
				{
					pack = new Container( 0x9B2, 0x3C, 0x48, new Rectangle2D( 44, 65, 142, 94 ) );//stuff for a regular backpack
					pack.Layer = Layer.ShopBuy;
					AddItem( pack );
				}

				return pack;
			}
		}

		public abstract void InitSBInfo();

		private void LoadSBInfo()
		{
			m_LastRestock = DateTime.Now;

			InitSBInfo();
			
			for ( int i = 0; i < SBInfos.Count; i++ )
			{
				SBInfo sbInfo = (SBInfo)SBInfos[i];
				m_ArmorBuyInfo.AddRange( sbInfo.BuyInfo );
				m_ArmorSellInfo.Add( sbInfo.SellInfo );
			}
		}

		public virtual void InitBody()
		{
			InitStats( 100, 100, 25 );

			SpeechHue = Utility.RandomDyedHue();
			Hue = Utility.RandomSkinHue();
			NameHue = 0x35;

			if ( this.Female = Utility.RandomBool() )
			{
				this.Body = 0x191;
				this.Name = NameList.RandomFemaleName();
			}
			else
			{
				this.Body = 0x190;
				this.Name = NameList.RandomMaleName();
			}
		}
		
		public virtual void InitOutfit()
		{
			AddItem( new FancyShirt( Utility.RandomNeutralHue() ) );
			AddItem( new ShortPants( Utility.RandomNeutralHue() ) );
			AddItem( new Boots( Utility.RandomNeutralHue() ) );

			Item hair = new Item( Utility.RandomList( 0x203B, 0x2049, 0x2048, 0x204A ) );

			hair.Hue = Utility.RandomNondyedHue();
			hair.Layer = Layer.Hair;
			hair.Movable = false;

			AddItem( hair );

			Container pack = new Backpack();

			pack.DropItem( new Gold( 1000, 2000 ) );

			pack.Movable = false;

			AddItem( pack );
		}

		public virtual void Restock()
		{
			m_LastRestock = DateTime.Now;

			IBuyItemInfo[] buyInfo = this.GetBuyInfo();

			foreach ( IBuyItemInfo bii in buyInfo )
				bii.OnRestock();
		}

		public virtual void VendorBuy( Mobile from )
		{
			if ( DateTime.Now - m_LastRestock > RestockDelay )
				Restock();

			int count = 0;
			ArrayList list;
			IBuyItemInfo[] buyInfo = this.GetBuyInfo();
			IShopSellInfo[] sellInfo = this.GetSellInfo();

			list = new ArrayList( buyInfo.Length );

			for (int idx=0;idx<buyInfo.Length;idx++)
			{
				IBuyItemInfo buyItem = (IBuyItemInfo)buyInfo[idx];
				if ( buyItem.Amount > 0 )
				{
					if ( list.Count < 250 )
					{
						list.Add( new BuyItemState( buyItem.Name, BuyPack.Serial, 0x7FFFFFFF - idx, buyItem.Price, buyItem.Amount, buyItem.ItemID, buyItem.Hue ) );
						count++;
					}
				}
			}

			Container cont = this.BuyPack;
			foreach ( Item item in cont.Items )
			{
				int price = 0;
				string name = "";

				foreach( IShopSellInfo ssi in sellInfo )
				{
					if ( ssi.IsSellable( item ) )
					{
						price = ssi.GetBuyPriceFor( item );
						name = ssi.GetNameFor( item );
						break;
					}
				}

				if ( list.Count < 250 )
				{
					list.Add( new BuyItemState( name, BuyPack.Serial, item.Serial, price, item.Amount, item.ItemID, item.Hue ) );
					count++;
				}
			}

			//one (not all) of the packets uses a byte to describe number of items in the list.  Osi = dumb.
			//if ( list.Count > 255 )
			//	Console.WriteLine( "Vendor Warning: Vendor {0} has more than 255 buy items, may cause client errors!", this );

			if ( list.Count > 0 )
			{
				list.Sort( new BuyItemStateComparer() );

				from.Send( new VendorBuyContent( list ) );
				from.Send( new VendorBuyList( this, list ) );
				from.Send( new DisplayBuyList( this ) );
				from.Send( new MobileStatusExtended( from ) );//make sure their gold amount is sent

				SayTo( from, 500186 ); // Greetings.  Have a look around.
			}
		}

		public virtual void VendorSell( Mobile from )
		{
			Container pack = from.Backpack;

			if ( pack != null )
			{
				IShopSellInfo[] info = GetSellInfo();

				Hashtable table = new Hashtable();

				foreach ( IShopSellInfo ssi in info )
				{
					Item[] items = pack.FindItemsByType( ssi.Types );

					foreach ( Item item in items )
					{
						if ( item.LootType == LootType.Regular && item.Movable && ssi.IsSellable( item ) )
							table[item] = new SellItemState( item, ssi.GetSellPriceFor( item ), ssi.GetNameFor( item ) );
					}
				}

				if ( table.Count > 0 )
				{
					from.Send( new VendorSellList( this, table ) );
				}
				else
				{
					Say( true, "You have nothing I would be interested in." );
				}
			}
		}

		public virtual bool OnBuyItems( Mobile buyer, ArrayList list )
		{
			IBuyItemInfo[] buyInfo = this.GetBuyInfo();
			IShopSellInfo[] info = GetSellInfo();
			int totalCost = 0;
			ArrayList validBuy = new ArrayList( list.Count );
			Container cont;
			bool bought = false;
			bool fromBank = false;
			bool fullPurchase = true;
			int controlSlots = buyer.FollowersMax - buyer.Followers;

			foreach ( BuyItemResponse buy in list )
			{
				Serial ser = 0x7FFFFFFF - buy.Serial;
				int amount = buy.Amount;
				if ( ser >= 0 && ser <= buyInfo.Length )
				{
					IBuyItemInfo bii = buyInfo[ser];
					if ( amount > bii.Amount )
						amount = bii.Amount;
					if ( amount <= 0 )
						continue;

					int slots = bii.ControlSlots * amount;

					if ( controlSlots >= slots )
					{
						controlSlots -= slots;
					}
					else
					{
						fullPurchase = false;
						continue;
					}

					totalCost += bii.Price * amount;
					validBuy.Add( buy );
				}
				else
				{
					Item item = World.FindItem( buy.Serial );

					if ( item == null || item.RootParent != this )
						continue;

					if ( amount > item.Amount )
						amount = item.Amount;
					if ( amount <= 0 )
						continue;

					foreach( IShopSellInfo ssi in info )
					{
						if ( ssi.IsSellable( item ) )
						{
							if ( ssi.IsResellable( item ) )
							{
								totalCost += ssi.GetBuyPriceFor( item );
								validBuy.Add( buy );
								break;
							}
						}
					}
				}
			}//foreach

			if ( fullPurchase && validBuy.Count == 0 )
				SayTo( buyer, 500190 ); // Thou hast bought nothing!
			else if ( validBuy.Count == 0 )
				SayTo( buyer, 500187 ); // Your order cannot be fulfilled, please try again.

			if ( validBuy.Count == 0 )
				return false;

			cont = buyer.Backpack;
			if ( cont != null )
			{
				if ( cont.ConsumeTotal( typeof( Gold ), totalCost ) )
					bought = true;
				else if ( totalCost < 2000 )
					SayTo( buyer, 500192 );//Begging thy pardon, but thou casnt afford that.
			}

			if ( !bought && totalCost >= 2000 )
			{
				cont = buyer.BankBox;
				if ( cont.ConsumeTotal( typeof( Gold ), totalCost ) )
				{
					bought = true;
					fromBank = true;
				}
				else
				{
					SayTo( buyer, 500191 ); //Begging thy pardon, but thy bank account lacks these funds.
				}
			}

			if ( !bought )
				return false;
			else
				buyer.PlaySound( 0x32 );

			cont = buyer.Backpack;
			if ( cont == null )
				cont = buyer.BankBox;

			foreach ( BuyItemResponse buy in validBuy )
			{
				Serial ser = 0x7FFFFFFF - buy.Serial;
				int amount = buy.Amount;

				if ( ser >= 0 && ser <= buyInfo.Length )
				{
					IBuyItemInfo bii = buyInfo[ser];

					if ( amount > bii.Amount )
						amount = bii.Amount;

					bii.Amount -= amount;

					object o = bii.GetObject();

					if ( o is Item )
					{
						Item item = (Item)o;

						if ( item.Stackable )
						{
							item.Amount = amount;

							if ( !cont.TryDropItem( buyer, item, false ) )
								item.MoveToWorld( buyer.Location, buyer.Map );
						}
						else
						{
							item.Amount = 1;
							if ( !cont.TryDropItem( buyer, item, false ) )
								item.MoveToWorld( buyer.Location, buyer.Map );

							for (int i=1;i<amount;i++)
							{
								item = bii.GetObject() as Item;

								if ( item != null )
								{
									item.Amount = 1;
									if ( !cont.TryDropItem( buyer, item, false ) )
										item.MoveToWorld( buyer.Location, buyer.Map );
								}
							}
						}
					}
					else if ( o is Mobile )
					{
						Mobile m = (Mobile)o;

						m.Direction = (Direction)Utility.Random( 8 );
						m.Location = buyer.Location;
						m.Map = buyer.Map;
						m.PlaySound( m.GetIdleSound() );

						if ( m is BaseCreature )
							((BaseCreature)m).SetControlMaster( buyer );

						for ( int i = 1; i < amount; ++i )
						{
							m = bii.GetObject() as Mobile;

							if ( m != null )
							{
								m.Direction = (Direction)Utility.Random( 8 );
								m.Location = buyer.Location;
								m.Map = buyer.Map;

								if ( m is BaseCreature )
									((BaseCreature)m).SetControlMaster( buyer );
							}
						}
					}
				}
				else
				{
					Item item = World.FindItem( buy.Serial );

					if ( item == null )
						continue;

					if ( amount > item.Amount )
						amount = item.Amount;

					foreach( IShopSellInfo ssi in info )
					{
						if ( ssi.IsSellable( item ) )
						{
							if ( ssi.IsResellable( item ) )
							{
								Item buyItem;
								if ( amount >= item.Amount )
								{
									buyItem = item;
								}
								else
								{
									buyItem = item.Dupe( amount );
									item.Amount -= amount;
								}

								if ( !cont.TryDropItem( buyer, buyItem, false ) )
									buyItem.MoveToWorld( buyer.Location, buyer.Map );

								break;
							}
						}
					}
				}
			}//foreach

			if ( fullPurchase )
			{
				if ( fromBank )
					SayTo( buyer, true, "The total of thy purchase is {0} gold, which has been withdrawn from your bank account.  My thanks for the patronage.", totalCost );
				else
					SayTo( buyer, true, "The total of thy purchase is {0} gold.  My thanks for the patronage.", totalCost );
			}
			else
			{
				if ( fromBank )
					SayTo( buyer, true, "The total of thy purchase is {0} gold, which has been withdrawn from your bank account.  My thanks for the patronage.  Unfortunately, I could not sell you all the goods you requested.", totalCost );
				else
					SayTo( buyer, true, "The total of thy purchase is {0} gold.  My thanks for the patronage.  Unfortunately, I could not sell you all the goods you requested.", totalCost );
			}

			return true;
		}

		public virtual bool OnSellItems( Mobile seller, ArrayList list )
		{
			seller.PlaySound( 0x32 );

			IShopSellInfo[] info = GetSellInfo();
			IBuyItemInfo[] buyInfo = this.GetBuyInfo();
			int GiveGold = 0;
			int Sold = 0;
			Container cont;
			ArrayList delete = new ArrayList();
			ArrayList drop = new ArrayList();

			foreach ( SellItemResponse resp in list )
			{
				if ( resp.Item.RootParent != seller || resp.Amount <= 0 )
					continue;

				foreach( IShopSellInfo ssi in info )
				{
					if ( ssi.IsSellable( resp.Item ) )
					{
						Sold++;
						break;
					}
				}
			}

			if ( Sold > MaxSell )
			{
				SayTo( seller, true, "You may only sell {0} items at a time!", MaxSell );
				return false;
			} 
			else if ( Sold == 0 )
			{
				return true;
			}

			foreach ( SellItemResponse resp in list )
			{
				if ( resp.Item.RootParent != seller || resp.Amount <= 0 )
					continue;

				foreach( IShopSellInfo ssi in info )
				{
					if ( ssi.IsSellable( resp.Item ) )
					{
						int amount = resp.Amount;
						if ( amount > resp.Item.Amount )
							amount = resp.Item.Amount;

						if ( ssi.IsResellable( resp.Item ) )
						{
							bool found = false;
							foreach ( IBuyItemInfo bii in buyInfo )
							{
								if ( bii.Restock( resp.Item ) )
								{
									found = true;
									if ( resp.Item.Amount - resp.Amount <= 0 )
										resp.Item.Delete();
									break;
								}
							}

							//not in their buy list, but it is resellable, so put it in the resale pack
							if ( !found )
							{
								cont = this.BuyPack;
								if ( resp.Amount < resp.Item.Amount )
								{
									resp.Item.Amount -= resp.Amount;
									cont.DropItem( resp.Item.Dupe( resp.Amount ) );
								}
								else
								{
									cont.DropItem( resp.Item );
								}
							}
							else 
							{ 
								if ( resp.Amount < resp.Item.Amount ) 
									resp.Item.Amount -= resp.Amount; 
								else 
									resp.Item.Delete(); 
							}
						}
						else
						{
							resp.Item.Delete();
						}

						GiveGold += ssi.GetSellPriceFor( resp.Item )*amount;
						break;
					}
				}
			}

			if ( GiveGold > 0 )
			{
				Item gold;
				cont = seller.Backpack;
				if ( cont != null )
				{
					gold = cont.FindItemByType( typeof( Gold ) );
					if ( gold != null )
					{
						gold.Amount += GiveGold;
					}
					else 
					{
						gold = new Gold( GiveGold );
						if ( !cont.TryDropItem( seller, gold, true ) )
							gold.MoveToWorld( seller.Location, seller.Map );
					}
				}
				else
				{
					gold = new Gold( GiveGold );
					gold.MoveToWorld( seller.Location, seller.Map );
				}
				seller.PlaySound( 0x0037 );//Gold dropping sound
			}
			//no cliloc for this?
			//SayTo( seller, true, "Thank you! I bought {0} item{1}. Here is your {2}gp.", Sold, (Sold > 1 ? "s" : ""), GiveGold );
			
			return true;
		}

		public override void Serialize( GenericWriter writer )
		{
			base.Serialize( writer );

			writer.Write( (int) 0 ); // version
		}

		public override void Deserialize( GenericReader reader )
		{
			base.Deserialize( reader );

			int version = reader.ReadInt();

			LoadSBInfo();
		}

		public override void GetContextMenuEntries( Mobile from, ArrayList list )
		{
			base.GetContextMenuEntries( from, list );

			list.Add( new VendorBuyEntry( from, this ) );
			list.Add( new VendorSellEntry( from, this ) );
		}

		public virtual IShopSellInfo[] GetSellInfo()
		{
			return (IShopSellInfo[])m_ArmorSellInfo.ToArray( typeof( IShopSellInfo ) );
		}

		public virtual IBuyItemInfo[] GetBuyInfo()
		{
			return (IBuyItemInfo[])m_ArmorBuyInfo.ToArray( typeof( IBuyItemInfo ) );
		}
	}
}

namespace Server.ContextMenus
{
	public class VendorBuyEntry : ContextMenuEntry
	{
		private BaseVendor m_Vendor;

		public VendorBuyEntry( Mobile from, BaseVendor vendor ) : base( 6103, 8 )
		{
			m_Vendor = vendor;
		}

		public override void OnClick()
		{
			m_Vendor.VendorBuy( this.Owner.From );
		}
	}

	public class VendorSellEntry : ContextMenuEntry
	{
		private BaseVendor m_Vendor;

		public VendorSellEntry( Mobile from, BaseVendor vendor ) : base( 6104, 8 )
		{
			m_Vendor = vendor;
		}

		public override void OnClick()
		{
			m_Vendor.VendorSell( this.Owner.From );
		}
	}
}
